#!/store/bin/perl5 -w
#
# Author: Petter Reinholdtsen <pere@td.org.uit.no>
# Date:   1998-12-08
#
# Check libtool libraries to see if all symlinks are present.  Use to
# check if libraries are installed correctly in Store.
#
# Params: $0 <dir-list>

$prefix = "/store";

while (@ARGV) {
    my $dir = shift;
    unless (-d $dir) {
	print STDERR "No such directory: $dir\n";
	next;
    }
    for $file (<$dir/*.la>) {
	check_libtool_lib($file);
    }
}

exit 0;

sub check_libtool_lib ($) {
    my $filename = shift;

    unless ( open(LTF, "< $filename") ) {
	print STDERR "Unable to read $filename -> ",readlink($filename),"\n";
	return;
    }

    my @libs;
    my $dir;
    while (<LTF>) {
	@libs = split(/ /, $1) if (/^library_names='(.+)'$/);
	$dir = $1 if (/^libdir='(.+)'$/);
    }
    my @missing;
    for $lib (@libs) {
	my $fullname = "$dir/$lib";
	push(@missing, $fullname) unless (-f "$fullname");
	push(@wrong, $fullname) unless (readlink($filename) =~ m/$prefix/);
    }
    if (@missing) {
	print STDERR "$filename -> ",readlink($filename),"\n";
	for (@missing) {
	    print STDERR " Missing $_\n";
	}
	for (@wrong) {
	    print STDERR " Wrong $_\n";
	}
    }
}
