#!/usr/bin/perl
#
# Maintainer: Petter Reinholdtsen <pere@td.org.uit.no>
# Date:       1999-04-14

# check application consistency and send complaints to the person
# responsible for that application.
# -- bjoerns@acm.org 1995-12-12
#
# Changes:
#  - get list of master stores on the current machine automatically
#    [pere@td.org.uit.no 1998-11-16]

$prefix='/store' ;

unshift( @INC, "$prefix/etc/internal" ) ;
unshift( @INC, "." ) ;

require "lib/client.pl" ;
require "lib/conffile.pl" ;
require "lib/config.pl" ;
require "lib/genarchs.pl" ;
require "lib/indata.pl" ;
require "lib/master.pl" ;
require "lib/output.pl" ;
require "lib/regist.pl" ;
require "lib/registerit.pl" ;
require "lib/runit.pl" ;
require "lib/shadowit.pl" ;
require "lib/slaveit.pl" ;
require "lib/support.pl" ;
require "lib/utils.pl" ;
require "lib/notlinks.pl" ;

&'readgenarchs ;
&'readconffile("$config") ;

$'myself = &mymachine() ;

$'debug = $'WARNINGS ;

# Setup finished

# Find master stores on this machine
@allstores = &getstorelist ;
@myservers = grep( $'master{$_} eq $'myself , @allstores ) ;
@myservers = &'uniq( @myservers ) ;

# Try to find sendmail program
$sendmail = (-x "/usr/lib/sendmail" ) ? "/usr/lib/sendmail" : "/usr/sbin/sendmail";

for $server ( @myservers ) {
    @APPS = &'getmastapplist($server);

    $tmpPrefix = "/tmp/chkapp$$";


    # check all applications
    open(SAVEOUT, ">&STDOUT")	|| die "dup: $!";
    foreach $appl (@APPS) {
	&'readregist($server, $appl)	|| die "readregist: $!";
	next unless $'regist{'version'};

	$sig = $'regist{'signature'};
	($email) = ($sig =~ /<([^>]+)>/);

	if (!$email) {
		print STDERR "$appl contains erronous signature: $sig\n";
		next;
	}

#	print STDERR "reporting $appl to $email\n";
	open(STDOUT, ">>$tmpPrefix$email") || die "open($tmpPrefix$email): $!";

	$file{"$tmpPrefix$email"} = $email;

	&master'procapp($server, $appl, 0);
	# XXX: should also check that registration file is updated,
	# with a decent log, if there is a new version in the Store.
	if ($'regist{'sec_description'} !~ /\S/) {
		print STDERR "reporting $appl to $email\n";
		&'output('Warning', "empty description\n", '', $server, $appl);
	}
	close(STDOUT);
	open(STDOUT, ">&SAVEOUT");
    }
    close(SAVEOUT);

}

#while (<>) {
#	search for suid-files on master.sh output

print STDERR "***********************************\n";
print STDERR "Sending errors to contact persons\n";

# send reports to whoever is in charge
foreach $file (keys %file) {
	local($email) = $file{$file}; # ($file =~ /^$tmpPrefix(.*)/);

if (0) { # Send to everyone
	if ($email !~ m/(\@[^@]+.uit.no|pere\@|bjoerns\@|^[^@]+$)/) {
		print STDERR "non-local email address: $email, not sending\n";
		unlink($file);
		next;
	}
}
	if (-z $file) {
		unlink($file);
		next;
	}
	print STDERR "sending email to $email\n";
	
	open(MAIL, "|$sendmail $email")	|| die "|sendmail: $!";
##	open(MAIL, "|cat")	|| die "|sendmail: $!";
	print MAIL "To: $email\nFrom: Store's Nightly Check $'myself <store>\n"
		."Subject: Store errors you have to fix at $'myself\n\n";
	print MAIL "  [for  legge inn description;  logg in som store p $'myself og kjr\n   `register' i katalogen til applikasjonen (~ms/applikasjonsnavn).\n   du vil fortsette  f feilmeldinger hver natt inntil du fikser dem. ]\n\n";
	if (open(ERRORS, "$file")) {
		print MAIL join("", <ERRORS>);
		close(ERRORS);
	}
	if (open(ERRORS, "$file")) {
		print STDERR join("", <ERRORS>);
		close(ERRORS);
	}
	print STDERR "*********************************************\n";
	unlink($file);
	close(MAIL);
}
