# This is a -*- perl -*- script, but it isn't directly runnable.
# Say "perl scriptname" instead.
#
# It needs perl5 to work correctly!

require "getdir.pl";

# Make it a package
package Env;

$TOP = $'ENV{'TOPDIR'} || '/store';

$Indent = 4;
$Indent_step = 3;

##
## Do file globbing on path content.  Convert '/store/opt/*/bin' to
## all bin/ directories in /store/opt, and leave '/nonexistant/direcory'
## alone.
##
sub path_expand_fileglobbing {
    my $path = shift;
    my ($dir, @dirs);
    for $dir (split(/:/, $path)) {
	next unless ($dir);
        if ($dir =~ m/[\*\?]+/) {
	  eval "push \@dirs, <$dir>";
	} else {
	  push(@dirs, $dir);
	}
    }
    return join(':', @dirs);
}

##
## Read all environment settings in $dir
##
sub read_env {
    local( $dir ) = @_;
    return unless -d $dir;

    ## Check each file in dir
    local( @files ) = &'getAdir($dir);
    local( $file, $app, $t, $u, $g );
    local( $uid, $gid, $pri, $type, $env, $content, $uneg, $gneg );
    local( %allenv );
    foreach $file ( @files ) {
	## Filnames contain 'ENV-<identifier>'
	( $app ) = ( $file =~ /ENV-(.*)$/ );
	next unless $app;

	## Check specific file
	open( ENV, "$dir/$file" );
	while ( <ENV> ) {
	    ## Skip comments and blank lines
	    next if /^\#/;
	    next if /^\s*$/;

	    ## Things may be spanned over several lines using \.
	    while ( /\\$/ ) {
		s/\s*\\\s*$//;
		$t = <ENV>; $t =~ s/^\s+//;
		$_ .= ":$t";
	    }
	    chop;

	    ## Decode information for a single variable
	    ( $uid, $gid, $pri, $type, $env, $content ) = 
		/^(.*):(.*):(.*):(.):(\w+)\s*=\s*(.*)$/;
	    $content =~ s/\$TOP(DIR)?/$TOP/g;

	    ## Store one entry for each (uid,gid) - pair
	    foreach $u ( split( /,/, $uid ) ) {
		$uneg = ( $u =~ s/^\!// );
		$u =~ s/(^|,)0($|,)/\1root\2/g;
		## Get uid if available
		unless ( $u =~ /\w|^\*$/ ) {
		    if ( $uid{$u} ) { $u = $uid{$u}; }
		    elsif ( $t = getpwuid( $u ) ) { $u = $uid{$u} = $t; }
		}
		$u = "!$u" if $uneg;

		foreach $g ( split( /,/, $gid ) ) {
		    $gneg = $g =~ s/^!//;
		    ## Get gid if available		  
		    unless ( $g =~ /\w|^\*$/ ) {
			if ( $gid{$g} ) { $g = $gid{$g}; }
			elsif ( $t = getgrgid( $g ) ) { $g = $gid{$g} = $t; }
		    }
		    $g = "!$g" if $gneg;

		    ## Store it
		    if ($allenv{$env,$pri,$u,$g,$type,$app}) {
		        $allenv{$env,$pri,$u,$g,$type,$app} .= ":".$content;
		    } else {
		        $allenv{$env,$pri,$u,$g,$type,$app} = $content;
		    }
		}
	    }
	}
	close ENV;
    }

    ## Remove unwanted entries
    local( $p, %has_sv_env, %know_env, @envlist, $k, $key );
    local( %uids_included, %uid_is_included );
    local( %gids_included, %gid_is_included );
    foreach $k ( sort keys %allenv ) {
	( $env, $pri, $uid, $gid, $type, $app ) = split( /$;/, $k );

	if ( $type eq 's' ) {
	    ## Delete low-pri entries for single valued variables
	    delete $allenv{$k}, next if $has_set_env{$env,$uid,$gid} ||
		$has_set_env{$env,'*','*'};
	    $has_set_env{$env,$uid,$gid} = 1;
	    ## Make single valued vars be set very last
	    $uid =~ tr/*/~/;  $gid =~ tr/*/~/;
	    $allenv{$env,"~$env",$uid,$gid,$type,$app} = $allenv{$k};
	    delete $allenv{$k};

	} elsif ( $type eq 'p' ) {
	    ## Bulid list of all path-like variables in use
	    ## Delete non-existing directories
	    $allenv{$k} =
  join(':', grep( -d, split( /:/, &path_expand_fileglobbing($allenv{$k}) ) ) );
	    if ($allenv{$k}) {
	      ## Make a list of all path-variables
	      push( @envlist, $env ) unless $know_env{$env};
	      $know_env{$env} = 1;
	    } else {
	      delete $allenv{$k};
	      next;
	    }
	} elsif ( $type eq 'n' ) {
	    ## Make a list of all path-variables
	    push( @envlist, $env ) unless $know_env{$env};
	    $know_env{$env} = 1;
#	    print "[$env][$type][$allenv{$k}]\n";
	    ## From now on, non-path-checking variables should look
	    ## like normal path-variables
	    $key = join( $;, $env, $pri, $uid, $gid, 'p', $app );
	    $allenv{$key} = $allenv{$k};
	    delete $allenv{$k};
	}
    }

    ## Collect first assignment of a variable that everyone should
    ## know to the beginning of the list (set priority to 0).
    foreach $env ( @envlist ) {
	$last_change_key{$env} = join( $;, '+', '*', '*', $env, 'p');
	$last_change_id{$env} = "*:*";
	$prereset_env{$env} = 1;
    }

    ## Concatenate several priorities if possible
    local( $key );
    foreach $k ( sort keys %allenv ) {
	( $env, $pri, $uid, $gid, $type, $app ) = split( /$;/, $k );
	if ( $last_change_id{$env} eq "$uid:$gid" ) {
	    ## Concatenate
	    $key = $last_change_key{$env};
	    $env{$key} .= $env{$key} ? ":$allenv{$k}" : $allenv{$k};
  
	} else {
	    ## Stop concatenation
	    $env{$pri,$uid,$gid,$env,$type} = $allenv{$k};
	    $last_change_key{$env} = join($;,$pri,$uid,$gid,$env,$type);
	}
	delete $prereset_env{$env} if $env{'+','*','*',$env,$type};
	$last_change_id{$env} = "$uid:$gid";
    }

    ## Environment that must be reset prior to assigning
    foreach $env ( keys %prereset_env ) {
	$env{'*','*','*',$env,'p'} = '';
    }


    ## Make comma-separeted lists instead of single values for uids
    ## and gids (eg.  <uid 1:gid 10> and <uid 1: gid 14> is translated
    ## into <uid 1:gid 10,14> ).
    local( $val, %uid_list, %gid_list, %has_added_uid, %has_added_gid );
    foreach $k ( keys %env ) {
	( $pri, $uid, $gid, $env, $type ) = split( /$;/, $k );
	$val = $env{$k};
	delete $env{$k} if $uid_list{$env,$val,$pri} ne '';
	$uid_list{$env,$val,$pri} .= $uid_list{$env,$val,$pri} ? ",$uid":$uid
	    unless $has_added_uid{$env,$val,$pri,$uid};
	$gid_list{$env,$val,$pri} .= $gid_list{$env,$val,$pri} ? ",$gid":$gid
	    unless $has_added_gid{$env,$val,$pri,$gid};
	$has_added_uid{$env,$val,$pri,$uid} = 
	    $has_added_gid{$env,$val,$pri,$gid} = 1;
    }
    foreach $k ( keys %env ) {
	( $pri, $uid, $gid, $env, $type ) = split( /$;/, $k );
	$val = $env{$k};
	if ($uid_list{$env,$val,$pri} ne $uid ||
	    $gid_list{$env,$val,$pri} ne $gid ) {
	    delete $env{$k};
	    $env{$pri,$uid_list{$env,$val,$pri},
		 $gid_list{$env,$val,$pri},$env,$type} = $val;
	}
    }

    ## Return a nice assoc-array with elements
    ## { uid-list, gid-list, env } = value
    return %env;
}


##
## Indent $lines and return them, indenting by $pre levels before
## inserting, and $post levels after inserting.
##
sub insert {
    local( $pre, $lines, $post ) = ( @_, 0 );
    local( $str, $ind );
    
    $Indent += ( $pre * $Indent_step );
    $ind = ' ' x $Indent;
    $lines = $ind . $lines;
    $lines =~ s/\n(.)/\n$ind$1/g;
    $Indent += ( $post * $Indent_step );

    $lines;
}


##
## Build the setting of an environment-variable, $env with value
## $val, restricting the width to $width (if possible).  Result
## is in bourne-sh syntax.
##
sub build_envset {
    local( $env, $val, $width ) = ( @_, 80 );
    local( $str, $c, $al, $v ) = ( "$env=", 0, 2+length( $env ) );
    $width -= ($al-1);

    local( $l );
    foreach $v ( split( /:/, $val ) ) {
	$l = length( $v );
	if ( $c+$l <= $width || $l > $width ) {
	    $str .= $c == 0 ? $v : ":$v";
	    $c += $l;
	} else {
	    $str .= "\n$env=\$$env:$v";
	    $c = $l + $al;
	}
    }

    "$str\n";
}


##
## Build test string to be used in if-statements
## (bourne-shell syntax)
##
sub build_test {
    local( $str, $list ) = @_;
    local( $test, $t );

    foreach $t ( split( /,/, $list ) ) {
	$test .= "[ \"x$str\" = 'x$t' ]";
    }
    $test =~ s/\]\[/\] || \[/g;
    $test =~ s/= \'!/!= \'/g;
    $test =~ s/(!=[^\]]+\]) \|\|/\1 &&/g;

    $test
}

##
## The header of every file
sub intro {
    return <<EOF;
# Do not edit this file.  It is automatically created by the Store
# application env-config based on the content of /store/etc/ENV/ENV-*
# 
EOF
}

##
## Build bourne shell code that will set the correct
## environment based on uid and grouplist.
##
sub build_env {
    local( %env ) = @_;
    local( $str );

    $str .= &intro();

    $str .= &insert( 0, '__OLDPATH="$PATH"'."\n" );
    $str .= &insert( 0, "PATH=\"/usr/bin:/usr/ucb\"\n\n" );
    $str .= &insert( 0, "__UID=`whoami`\n" );
    $str .= &insert( 0, "__GIDS=`/usr/bin/groups`\n\n" );
    $str .= &insert( 0, "PATH=\"\$__OLDPATH\"\n" );
    $str .= &insert( 0, "unset __OLDPATH\n" );

    $str .= &insert( 0, "\n# Make sure environment is exported\n" );
    $str .= &insert( 0, "set -a\n\n" );

    local( $pri, $uid, $gid, $env, $type, $k );
    local( $envset, $anvadd, $test, $neg );
    local( %did_pri );
    local( $ouid, $ogid, $oenv );
    foreach $k ( sort keys %env ) {
	( $pri, $uid, $gid, $env, $type ) = split( /$;/, $k );
	$envset = &build_envset( $env, $env{$k}, 66 );
	$envadd = &build_envset( $env, "\$$env\${$env:+:}$env{$k}", 66 );

	if ( $pri lt '0' ) {
	    $str .= &insert( 0, "## Reset some variables\n" )
		if !$did_pri{'*'} && $pri eq '*';
	    $str .= &insert( 0, "\n## Path variables\n" )
		if !$did_pri{'+'} && $pri eq '+';
	    $str .= &insert( 0, "$envset" );

	} elsif ( $pri =~ '^~' ) {
	    if ( $in_uid && !$did_pri{'~'} ) {
		$str .= &insert( -1, "fi\n" );
		$in_uid = 0;
	    }

	    $str .= &insert( 0, "\n\n## Other variabels\n" )
		unless $did_pri{'~'};
	    $uid =~ tr/~/*/; $gid =~ tr/~/*/;

	    if ( $oenv ne $env ) {
		$str .= &insert( -1, "fi\n" ) if $in_uid;
		$in_uid = 0;
		$test = &build_test( '$__UID', $uid );
		$str .= &insert( 0, "if $test\nthen\n", 1 ), $in_uid = 1
		    unless $uid eq '*';
	    } else {
		if ( $uid ne $ouid ) {
		    if ( $uid eq '*' ) {
			$str .= &insert( -1, "else\n", 1 );
		    } else {
			$test = &build_test( '$__UID', $uid );
			$str .= &insert( -1, "elif $test\nthen\n", 1 );
		    }
		}
	    }
	    if ( $gid ne $ogid ) {
		if ( $gid ne '*' ) {
		    $str .= &insert( 0, "for __GID in \$__GIDS\ndo\n", 1 );
		    $test = &build_test( '$__GID', $gid );
		    $str .= &insert( 0, "if $test\nthen\n", 1 );
		    $str .= &insert( 0, "${envset}break\n", -1 );
		    $str .= &insert( 0, "fi\n", -1 );
		    $str .= &insert( 0, "done\n" );
		} else {
		    $str .= &insert( 0, $envset );
		}
	    } else {
		$str .= &insert( 0, $envset );
	    }

	    $did_pri{'~'} = 1;
	    $oenv = $env;
	    
	} else {

	    if ( !$did_pri{$pri} || $uid ne $ouid || $gid ne $ogid ) {
		$str .= &insert( 0, "\n" ) unless $in_uid;
		if ( $in_uid ) {
		    $str .= &insert( -1, "fi\n\n" );
		    $in_uid = 0;
		}
	    }

	    if ( $uid ne $ouid ) {
		if ( $uid ne '*' ) {
		    $test = &build_test( '$__UID', $uid );
		    $str .= &insert( 0, "if $test\nthen\n", 1 );
		    $in_uid = 1;
		}
		if ( $gid eq '*' ) {
		    $did_pri{$pri} = 1;
		    $str .= &insert( 0, $envadd );
		    $ouid = $uid;  $ogid = $gid;
		    next;
		}
	    }

	    if ( $gid ne $ogid ) {
		if ( $gid ne '*' ) {
		    $str .= &insert( 0, "for __GID in \$__GIDS\ndo\n", 1 );
		    $test = &build_test( '$__GID', $gid );
		    $str .= &insert( 0, "if $test\nthen\n", 1 );
		    $str .= &insert( 0, "${envadd}break\n", -1 );
		    $str .= &insert( 0, "fi\n", -1 );
		    $str .= &insert( 0, "done\n" );
		} else {
		    $str .= &insert( 0, $envadd );
		}
	    }
	}
	$did_pri{$pri} = 1;
	$ouid = $uid;  $ogid = $gid;
    }

    $str .= &insert( -1, "fi\n" ) if $in_uid;

    $str .= &insert( 0, "\n# Stop exporting environment\n" );
    $str .= &insert( 0, "set +a\n" );

    $str .= &insert( 0, "\nunset __UID\n" );
    $str .= &insert( 0, "unset __GID\n" );
    $str .= &insert( 0, "unset __GIDS\n" );

    $str
}


##
## Build test string to be used in if-statements
## (cshell syntax)
##
sub build_csh_test {
    local( $str, $list ) = @_;
    local( $test, $t );

    foreach $t ( split( /,/, $list ) ) {
	$test .= "($str == '$t')";
    }
    $test =~ s/\)\(/\) || \(/g;
    $test =~ s/== \'!/!= \'/g;
    $test =~ s/(!=[^\)]+\)) \|\|/\1 &&/g;

    $test
}


##
## Build csh code that will set the correct
## environment based on uid and grouplist.
##
sub build_csh_env {
    local( %env ) = @_;
    local( $str );

    $str .= &intro();

    $str .= &insert( 0, "setenv __OLDPATH \"\$PATH\"\n" );
    $str .= &insert( 0, "setenv PATH \"/usr/bin:/usr/ucb\"\n" );
    $str .= &insert( 0, "set __UID=`/usr/bin/whoami`\n" );
    $str .= &insert( 0, "set __GIDS=(`/usr/bin/groups`)\n\n" );
    $str .= &insert( 0, "setenv PATH \"\$__OLDPATH\"\n" );
    $str .= &insert( 0, "unset __OLDPATH\n" );

    local( $pri, $uid, $gid, $env, $type, $k );
    local( $envset, $anvadd, $test, $neg );
    local( %did_pri );
    local( $ouid, $ogid, $oenv );
    local( $in_gid, $in_uid, %fix_env );
    foreach $k ( sort keys %env ) {
	( $pri, $uid, $gid, $env, $type ) = split( /$;/, $k );

	if ( $type eq 's' ) {
	    if ( $env{$k} ) { $envset = "setenv $env $env{$k}\n"; }
	    else { $envset = "setenv $env ''\n"; }
	} else {
	    if ( $env{$k} ) {
		$envset = "set i_$env=(". join(' ',split(/:/,$env{$k})). ")\n";
	    } else {
		$envset = "set i_$env\n";
	    }	    
	    $fix_env{$env} = 1;
	}
	$envadd = "set i_$env=(\$i_$env " .
	    join(' ',split(/:/,$env{$k})) . ")\n";


	if ( $pri lt '0' ) {
	    $str .= &insert( 0, "## Reset some variables\n" )
		if !$did_pri{'*'} && $pri eq '*';
	    $str .= &insert( 0, "\n## Path variables\n" )
		if !$did_pri{'+'} && $pri eq '+';
	    $str .= &insert( 0, "$envset" );

	} elsif ( $pri =~ '^~' ) {
	    if ( $in_uid && !$did_pri{'~'} ) {
		$str .= &insert( -1, "endif\n" );
		$in_uid = 0;
	    }

	    $str .= &insert( 0, "\n\n## Other variabels\n" )
		unless $did_pri{'~'};
	    $uid =~ tr/~/*/; $gid =~ tr/~/*/;

	    if ( $oenv ne $env ) {
		$str .= &insert( -1, "endif\n" ) if $in_uid;
		$in_uid = 0;
		$test = &build_csh_test( '$__UID', $uid );
		$str .= &insert( 0, "if ($test) then\n", 1 ), $in_uid = 1
		    unless $uid eq '*';
	    } else {
		if ( $uid ne $ouid ) {
		    if ( $uid eq '*' ) {
			$str .= &insert( -1, "else\n", 1 );
		    } else {
			$test = &build_csh_test( '$__UID', $uid );
			$str .= &insert( -1, "else if ($test) then\n", 1 );
		    }
		}
	    }
	    if ( $gid ne $ogid ) {
		if ( $gid ne '*' ) {
		    $str .= &insert( 0, "foreach __GID (\$__GIDS)\n", 1 );
		    $test = &build_csh_test( '$__GID', $gid );
		    $str .= &insert( 0, "if ($test) then\n", 1 );
		    $str .= &insert( 0, "${envset}break\n", -1 );
		    $str .= &insert( 0, "endif\n", -1 );
		    $str .= &insert( 0, "end\n" );
		} else {
		    $str .= &insert( 0, $envset );
		}
	    } else {
		$str .= &insert( 0, $envset );
	    }

	    $did_pri{'~'} = 1;
	    $oenv = $env;
	    
	} else {

	    if ( !$did_pri{$pri} || $uid ne $ouid || $gid ne $ogid ) {
		$str .= &insert( 0, "\n" ) unless $in_uid;
		if ( $in_uid ) {
		    $str .= &insert( -1, "endif\n\n" );
		    $in_uid = 0;
		}
	    }

	    if ( $uid ne $ouid ) {
		if ( $uid ne '*' ) {
		    $test = &build_csh_test( '$__UID', $uid );
		    $str .= &insert( 0, "if ($test) then\n", 1 );
		    $in_uid = 1;
		}
		if ( $gid eq '*' ) {
		    $did_pri{$pri} = 1;
		    $str .= &insert( 0, $envadd );
		    $ouid = $uid;  $ogid = $gid;
		    next;
		}
	    }

	    if ( $gid ne $ogid ) {
		if ( $gid ne '*' ) {
		    $str .= &insert( 0, "foreach __GID (\$__GIDS)\n", 1 );
		    $test = &build_csh_test( '$__GID', $gid );
		    $str .= &insert( 0, "if ($test) then\n", 1 );
		    $str .= &insert( 0, "${envadd}break\n", -1 );
		    $str .= &insert( 0, "endif\n", -1 );
		    $str .= &insert( 0, "end\n" );
		} else {
		    $str .= &insert( 0, $envadd );
		}
	    }
	}
	$did_pri{$pri} = 1;
	$ouid = $uid;  $ogid = $gid;
    }

    $str .= &insert( -1, "endif\n" ) if $in_uid;


    $str .= &insert( 0, "\n## Hack for setting path-variables\n" );
    foreach $env ( sort keys %fix_env ) {
	next if $env eq 'PATH';
	$str .= &insert( 0, "set path=(\$i_$env); " .
			"unset \$i_$env; " .
			"setenv $env \$PATH\n" );
    }
    if ( $fix_env{'PATH'} ) {
	$str .= &insert( 0, "set path=(\$i_PATH); unset \$i_PATH\n" );
    } else {
	$str .= &insert( 0, "set path\n" );
    }
    
    $str .= &insert( 0, "\nunset __UID\n" );
    $str .= &insert( 0, "unset __GID\n" );
    $str .= &insert( 0, "unset __GIDS\n" );

    $str
}


## Return a non-nil value
1;
