# This is a -*- perl -*- script, but it isn't directly runnable.
# Say "perl scriptname" instead.

# Author: Petter Reinholdtsen <pere@td.org.uit.no>
# Date:   1999-08-12
#
# Generate GNU info/dir file from START-INFO-DIR-ENTRYs in the .info
# files.  Based on a script on unknown NTNU origin.
#
# It expect all files to have entries like this:
#
#	INFO-DIR-SECTION Programming
#	START-INFO-DIR-ENTRY
#	* Cpp: (cpp).		       The GNU C preprocessor.
#	END-INFO-DIR-ENTRY
#
# ChangeLog:
#  1999-08-15	Rewrote to handle more then one entry in each file

use strict;
use vars qw($TOP $gunzip $diff
	    $infodir $section %sections %entries
	    $op $di $printFailed);

die "(emacs-info) Only runnable by store user!\n" if ($< != 52) ;
	
$TOP = $ENV{"TOPDIR"} ;

# We require GNU gzip to extract info.gz files
$gunzip = "gunzip";

# We require a diff which handles '-c' (context diff)
$diff = "diff";

scan_infodir("$TOP/info");

$printFailed = 0;
$op = "(emacs-info) Making emacs info dir file...\n";
unlink("dir.old");
rename("dir", "dir.old");
unlink("dir");
unless (open(DIROUT, ">dir")) { 
    print STDERR "(emacs-info) Unable to open $TOP/info/dir for writing\n";
    $printFailed = 1;
};
chmod(0644, "dir");

unless ($printFailed) {
    print_header();
    for $section (sort keys %sections) {
	unless (print DIROUT "\n$section\n") { $printFailed = 1};
	unless (print DIROUT $sections{$section}) { $printFailed = 1};
    }
}

close(DIROUT);

if ($printFailed) {
       unlink ("dir");
       rename ("dir.old", "dir");
       print STDERR "(emacs-info) Unable to write to file. Old file restored.\n";
       exit 1 ;
}

$op .= "(emacs-info) differences $TOP/info/dir.old $TOP/info/dir:\n" ;
$di  = `$diff -c dir.old dir` ;
$op .= $di ;
$op .= "(emacs-info) End dir\n" ;

print $op unless (( $di =~ /No differences/ ) || ( $di eq "" )) ;

unlink ("dir.old") ;
exit 0 ;

sub add_dir_entry {
  my ($section, $entry, $filename) = @_;
  if ($entries{$entry}) {
    print STDERR "(emacs-info) $$filename ".
      "duplicates entry in $entries{$entry}\n";
  } else {
    $entries{$entry} = "$filename";
    $sections{$section} .= $entry;
  }
}

sub scan_infodir {
  my $infodir = shift;
  my $file;
  chdir $infodir || die "unable to chdir to $infodir";
  for $file (<*.info *.info.gz>) {
    unless ( extract_dir_entry($file) ) {
      print STDERR "(emacs-info) $infodir/$file ".
	"missing START-INFO-DIR-ENTRY\n";
    }
  }
}

##
# Parse GNU info file, locating and extracting the INFO-DIR entry.
sub extract_dir_entry {
    my $filename = shift;
    if ($filename =~ /.gz$/) {
	open(INFOFILE, "$gunzip < $filename |") || return;
    } else {
	open(INFOFILE, "< $filename") || return;
    }
    my $direntry = 0;
    my $section = "";
    my $entry = "";
    my $count = 0;
    while (<INFOFILE>) {
      if (/^END-INFO-DIR-ENTRY/) {
	chomp $section;
	$count++;
	add_dir_entry($section, $entry, $filename);
	$direntry = 0;
	$section = "";
	$entry = "";
      }
      $section = $1 if /INFO-DIR-SECTION (.*)$/;
      $entry .= $_ if $direntry;
      $direntry = 1 if /^START-INFO-DIR-ENTRY/;
    }
    close(INFOFILE);
    return $count;
}

sub print_header {
    unless (print DIROUT <<END_HEADER) { $printFailed = 1};
-*- Text -*-
Autogenerated using $0

This is the file .../info/dir, which contains the topmost node of the
Info hierarchy.  The first time you invoke Info you start off
looking at that node, which is (dir)Top.

File: dir	Node: Top	This is the top of the INFO tree
  This (the Directory node) gives a menu of major topics. 
  Typing "d" returns here, "q" exits, "?" lists all INFO commands, "h" 
  gives a primer for first-timers, "mTexinfo<Return>" visits Texinfo topic,
  etc.  (This does not apply if you read this via a web browser.)
  In Emacs, you can click mouse button 2 on a menu item or cross reference
  to select it.
  --- PLEASE ADD DOCUMENTATION TO THIS TREE. (See INFO topic first.) ---

* Menu: This is the top-level list of documents available.

END_HEADER
}
