# This is a -*- perl -*- script, but it isn't directly runnable.
# Say "perl scriptname" instead.
#
# Author: Petter Reinholdtsen <pere@td.org.uit.no>
# Date:   1998-11-25
#
# 'prognews' reads thru all registration for new log entries printing
# the last day entries.  Use it in crontab like this:
#    0 5 * * 1 /store/bin/perl5 /store/etc/internal/prognews.pl -d 7
# This should give you the last 7 days of logging every monday at 5:00

$prefix='/store' ;

unshift( @INC, "$prefix/etc/internal" ) ;
unshift( @INC, "." ) ;
 
require "lib/client.pl" ;
require "lib/conffile.pl" ;
require "lib/config.pl" ;
require "lib/genarchs.pl" ;
require "lib/indata.pl" ;
require "lib/master.pl" ;
require "lib/output.pl" ;
require "lib/regist.pl" ;
require "lib/registerit.pl" ;
require "lib/runit.pl" ;
require "lib/shadowit.pl" ;
require "lib/slaveit.pl" ;
require "lib/support.pl" ;
require "lib/utils.pl" ;
require "lib/notlinks.pl" ;

use Time::Local;

&'readgenarchs ;
&'readconffile("$config") ;

$'myself = &mymachine() ;

$'debug = $'WARNINGS | $'INFO ;
$'debug = $'WARNINGS | $'INFO | $'TRACE if (-t) ;

# Setup finished

# Options:
#  -d days

# Main program

$daylimit = 1; # day

while ( $_ = $ARGV[0] ) {
    shift;
    $err=1, last if /^-../ ;
    if (/^-d/)  { $daylimit = shift ; next ; } ;
    $err=1; last ;
}
if ( $err ) {
   print "Usage: prognews [ -d days ]\n" ;
   exit 1;
}

@allstores = &getstorelist ;

@myservers = grep( $'master{$_} eq $'myself , @allstores ) ;
@myservers = &'uniq( @myservers ) ;

for $server ( @myservers ) {
    @applist = &'getmastapplist($server);
    for $appl ( @applist ) {
	&'readregist($server, $appl);
        $log = $'regist{'sec_log'};
        next unless ($log =~ m/\S/);
        parse_store_log($server, $appl, $log);
    }
}

sub parse_store_log {
    my ($server, $appl, $log) = @_;

    @lines = split(/\n/, $log);

    $timelimit = time()-$daylimit*24*60*60;

    my $dontprint = "1";
    for $line (@lines) {
	if ($dontprint) {
	    if ($line =~ m/LOGNAME=/) {
		$date = extractdate($line) ;
		if ($date > $timelimit) {
		    $dontprint = "";
		    print "\n$server/$appl\n";
		    print "  $line\n";
		}
	    }
	} else {
	    print "  $line\n" if ($line =~ /\S/);
	}
    }
}

sub extractdate
# Get line like this:
#  getlogin=, uid=store, LOGNAME=store  Fri Sep 18 13:16:05 CEST 1998
# and return seconds since epoc for the date (UTC)
# XXX This should be rewritten to use Date::Manip or equivalent
{
    my $line = shift;
#print "L: $line\n";
    %monmap = (Jan => 0, Feb => 1, Mar => 2, Apr => 3, May => 4, Jun
	       => 5, Jul => 6, Aug => 7, Sep => 8, Oct => 9, Nov =>
	       10, Dec => 11);

    $line =~ s/^.+LOGNAME=\S+\s+(.*)$/$1/;
    my ($mon, $mday, $hour, $min, $sec, $year) = 
	$line =~ m/\S+ *(\S+) *(\d+) *(\d+):(\d+):(\d+) +\D* *(\d+)$/;
    unless ($mon) {
	print "Unhandled date: $line\n";
	return 0;
    }

    $mon = $monmap{$mon};
    return timelocal($sec, $min, $hour, $mday, $mon, $year);
}
