# This is a -*- perl -*- script, but it isn't directly runnable.
# Say "perl scriptname" instead.
#
# $Id: sourcewatch.pl,v 1.17 1996/01/29 10:08:52 ssb Exp $
#
# This script looks for new versions of source for applications that
# have their Source field in the registration file set to a special
# format.  Sourcewatch will use ftp to see if there are any files with
# apparently higher version numbers and report the newest of these on
# stdout.

#
# Stig Sther Bakken (ssb@bibsys.no)
#

#
# TODO:
#
# * stresstesting:
#   - forking
#   - version number sorting
# * need a timeout for ftp sessions
#

# @INC and require

$prefix='/store' ;

unshift( @INC, "$prefix/etc/internal" ) ;
unshift( @INC, "." ) ;

require 'lib/conffile.pl' ;
require 'lib/config.pl' ;
require 'lib/genarchs.pl' ;
require 'lib/indata.pl' ;
require 'lib/output.pl' ;
require 'lib/regist.pl' ;
require 'lib/utils.pl' ;

require 'getopts.pl';

$'debug = $'WARNINGS | $'ERRORS | $'FATAL | $'INFO |
    (($'TRACE) * $'isinteractive);

#$'debug = $'WARNINGS | $'ERRORS | $'FATAL | $'INFO;

#$'debug = $'EVERY ;

# command-line parameters

&Getopts('a:cDde:F:fg:Hhm:Ssuv');

$'debug = ($'FATAL | $'ERRORS | $'WARNINGS) if $opt_s;
$'debug = $'NODEBUG if $opt_S;

if ($opt_D) {
    $'isinteractive = 0;
}

if ($opt_c && !$opt_u) {
    &'output('Warning', "-c option is meaningless without -u option.\n",
	     'SourceWatch');
}

#$master = $opt_m;

$'debug |= ($'TRACE | $'XDEBUG | $'TIMING) if $opt_d;
#$'debug = $'EVERY if $opt_d;

# global variables

undef %commands_by_apps;
undef %apps_by_sites;
undef %part1;
undef %part2;

undef $logfile;		# ftp session log file.  it is global so the signal
			# handler can remove the file when we are killed.

undef @Apps;		# array of applications to look for


# some defaults
$getcmd = $opt_g || 'ncftp';
$max_fork = $opt_F || 4;

if ($opt_e) {
    $ftp_email = $opt_e;
}
else {
    open( INITS, "$'etcdir/initials" ) ;
    for ( <INITS> ) {
        ( $username, $init, $email, $where, @fullname ) = split ;
        $admins{$username} = $init ;
        $uadmins{$init} = $username ;
        $admins{'email',$username} = $email ;
        $admins{'where',$username} = $where ;
        $admins{'fullname',$username} = "@fullname" ;
    }
    close( INITS ) ;
    $init = $admins{getlogin()};
    $init = $admins{(getpwuid($<))[0]} if( $init eq '' || $init eq 'store' );
    $init = $admins{(getpwuid($>))[0]} if( $init eq '' || $init eq 'store' );
    $init = $admins{$ENV{'LOGNAME'}}   if( $init eq '' || $init eq 'store' );
    $ftp_email = $admins{'email',$uadmins{$init}};
}

if ($opt_h) {
    die <<EOF;
Usage: sourcewatch [-a apps] [-c] [-D] [-d] [-e address] [-f] [-F procs]
                   [-g getcmd] [-H] [-h] [-m master] [-s] [-S] [-u] [-v]

Options:
  -a <appls>  Whitespace-separated list of applications to look for.
  -c          Show commands.  Has no effect when not printing URLs.
  -D          Use default values (don't prompt).
  -d          More debug output than you'd possibly ever want.
  -e          Email address to use when logging on anonymous ftp servers.
  -f          Fork off ftp processes in the background.  See -F.
  -F          Set the maximum number of forked processes at once.  The
              default value is $max_fork.
  -g <cmd>    Get command to download URLs.  The default value is "$getcmd".
  -H          Long help (explains how it works).
  -h          Short help (this message).
  -m <store>  The master store to use.  The default is to use all the
              master stores on the host we are running on.
  -s          Be silent.  Outputs only warnings, errors and fatal errors.
  -S          Be really silent.  For use with eval \`sourcewatch\` and such.
  -u          Print URLs.
  -v          Displays more than one found version (verbose output).

EOF
}

if ($opt_H) {
    die <<EOF;

                           SourceWatch
                           ===========

This program looks around on ftp sites for new versions of software
installed in Store.  To be able to use this program fully, you should
set up the source location field in the registration file to an ftp
address with ${version} in it.  You may give the address in this field
as an URL (FTP and HTTP are currently the only protocols supported),
or in the form ftp.site:/dir/file.  Note that HTTP works really bad
since sourcewatch can not do the equivalent of "dir file-*.tar.gz", so
you are advised to stick to FTP.  However, if you do use HTTP, the
${version} support goes away and the whole URL will be recommended for
downloading every time.  HTTP fields will not be shown if sourcewatch
is not running in print URL mode.

EOF
}

# Store setup

&'readgenarchs ;
&'readconffile("$config") ;
$'myself = &'mymachine() ;

# interactive parameters

if (!$opt_a) {
    if ($'isinteractive) {
	@Apps = split(/\s+/, &'simple("Which application(s)?", ''));
    }
    else {
	&'output('Trace', "Looking up applications on $'myself...\n", $$);
	if (!$utils'AppInfoInitialized) {
	    &'initappinfo;
	}
	if ($opt_m) {
	    @Apps = split("$;", $utils'mastapps{$opt_m});
	}
	else {
	    @Apps = ();
	    undef $str, $host;
	    while (($str, $host) = each %'master) {
		if ($'myself eq $host) {
		    push(@Apps, split("$;", $utils'mastapps{$str}));
		}
	    }
	    @Apps = &'uniq(@Apps);
	}
    }
}
else {
    @Apps = split(/\s+/, $opt_a);
}

#print "apps=@Apps\n";

if ($'debug & $'TIMING) {
    $time = time;
    @times = times;
}
foreach (&add_files_for_checking(@Apps, $master)) {
    &recommend($_);
}
if ($'debug & $'TIMING) {
    $time2 = time;
    @times2 = times;
    $et = $time2 - $time;
    $st = $times2[1] - $times[0];
    $ut = $times2[0] - $times[0];
    $s = sprintf("%ds elapsed, %.1fs user, %.1fs system.\n", $et, $ut, $st);
    &'output('Timing', "looking up source: $s", $$);
}


&'output('Trace', "Ready to check source locations.\n", $$);

# collect information on which versions exist out there

undef $site, $pid, @pids, %sites_by_pids, $child, $childsite;

@sitequeue = keys %apps_by_sites;

while (%apps_by_sites) {
    if ($opt_f) { 
        ##
        ## if we have at least as many "balls in the air" as we
        ## should, wait for processes to die before continuing.
	while (%sites_by_pids && $#pids+1 >= $max_fork) {
	    $child = waitpid(-1, 0);
	    $childsite = $sites_by_pids{$child};
	    if (!$childsite || !$apps_by_sites{$childsite}) {
		print "(Hilfe!) Etwas ist fehl! ($$)\n";
		&'dumpassoc("main'apps_by_sites");
	    }
	    else {
		delete $sites_by_pids{$child};
		delete $apps_by_sites{$childsite};
	    }
	}
	$site = shift @sitequeue;
        $pid = &handle_one_site($site, 1);
        $sites_by_pids{$pid} = $site;
    }
    else {
        &'output('Debug', "going to handle $site\n", $$);
        &handle_one_site($site, 0);
        delete $apps_by_sites{$site};
    }
    foreach $site (keys %apps_by_sites) {
        &'output('Debug', "going to handle $site\n", $$);
        $pid = &handle_one_site($site, $opt_f);
	##
	## handle_one_site will return undef if we are not forking.
	## in that case, just continue.
	if (!defined($pid)) {
	    delete $apps_by_sites{$site};
	    next;
	}
	##
	## if $pid has a value, we are the parent process.
	if ($pid) {
	    push(@pids, $pid);
	    $sites_by_pids{$pid} = $site;
	}
	##
	## if not we are the child process.
	else {
	    ##
	    ## get rid of the children so we don't need to worry more
	    ## about them. (sounding familiar? :-)
	    exit;
	    
	    ## if more code is to be added here, remember to undef all
	    ## variables like @pids, %sites_by_pids etc.
	}
    }
}

while (%sites_by_pids) {
    $child = waitpid(-1, 0);
    $childsite = $sites_by_pids{$child};
    if (!$childsite || !$apps_by_sites{$childsite}) {
	print "$$ STRANGENESS ERROR II!\n";
    }
    else {
	delete $sites_by_pids{$child};
	delete $apps_by_sites{$childsite};
    }
}

exit;


### Subroutines

sub recommend # ($app, [$ver, @vers])
{
    local ($app, $ver) = @_;
    local ($url, $path, $cmd);
    
    return undef unless $app;
    
    ## the -u option means print URLs
    if ($opt_u) {
	##
	## make life easy for those who use the -c option:
	## print a "cd" command so they know where the file should
	## be downloaded to.
	$cmd = '';
	if ($opt_c) {
	    $path = &'getappinfo($app, 'masterpath', $opt_m);
	    $cmd .= "cd $path\n";
	}
	if ($ver || @vers) {
	    @vers = ($ver) unless @vers;
	    foreach $ver (@vers) {
		$url = &'getappinfo($app, 'source', $opt_m);
		$url = &source_url($app, $ver, $opt_m) if $ver;
		next unless $url;
		$cmd .= "$getcmd " if $opt_c;
		$cmd .= "$url\n";
	    }
	}
	else {
	    $url = &'getappinfo($app, 'source', $opt_m);
	    $cmd .= "$getcmd " if $opt_c;
	    $cmd .= "$url\n";
	}

	# print everything at once so other processes can not get between
	# two prints.
	print $cmd;
    }
    else {
	if ($ver) {
	    print "$app $ver\n";
	}
	else {
	    if (@vers) {
		print "$app @vers\n";
	    }
	    else {
		print "$app *\n";
	    }
	}
    }
    1;
}


sub add_file_for_checking # ($appl, $site, $path)
#
# $path is of the form /foo/bar/gazonk-*.tar.gz
#
{
    local ($appl, $site, $path) = @_;
    local (@tmp, $file, $dir);
    
    @tmp = split("$;", $apps_by_sites{$site});
    push(@tmp, $appl);
    $apps_by_sites{$site} = join("$;", @tmp);
    &'output('Debug', "application added: $appl\n");
    $dir = &'dirname($path);
    $file = &'basename($path);
    $commands_by_apps{$app} = "cd $dir\ndir $file\n";
    
    ## this really requires that no source fields contains more than
    ## one wildcard.
    ( $part1{$app}, $part2{$app} ) = split(/\*/, $file);
    
    1;
}


sub check_site # ($site)
#
# this function probably only handles UNIX ftp servers.
#
{
    local ($site) = @_;
    local ($dir_commands, $app, $path, $dir, $file, @apps, $ftp_input,
	   $ftp_command, $expecting, $doing, $ver, @tmp,
	   %found_versions, $tmpdir);
    
    @apps = split("$;", $apps_by_sites{$site});
    return undef unless @apps;
    foreach (@apps) {
	$dir_commands .= $commands_by_apps{$_};
    }
    $ftp_input = "user anonymous -$ftp_email\n${dir_commands}bye";
    $ftp_command = "ftp -v -n $site << EOF\n$ftp_input\nEOF\n";
    open(FTP, "$ftp_command |") || &'output('Fatal',
					    "Could not run ftp.\n", $$);
    $tmpdir = $ENV{'TMPDIR'} || '/tmp';
    $logfile = "$tmpdir/ftplog-$site.$$";
    open(LOG, ">$logfile");
    select LOG; $| = 1; select STDOUT;
    print LOG "Log of ftp connection to $site started ", `date`;
    print LOG "> ", join("\n> ", split("\n", $ftp_input)), "\n";
    &'output('Dump',
	     "$site:\n > " . join("\n > ", split("\n", $ftp_input)) .
	     "\n", $$);
    &'output('Trace', " connection log in \"$logfile\".\n", $$);
    $expecting = shift @apps;
    while (<FTP>) {
	print LOG "< $_";
	&'output('Dump', "< $_", $$);
	chop;
	if ($expecting) {
	    ## wait until we are about to open a data connection.
	    next unless /^150 /;
	    next if /^total /;
	    $doing = $expecting;
	    $expecting = '';
	}
	elsif ($doing) {
	    ## 226 means the action was successful.
	    if (/^226 /) {
		$doing = '';
		$expecting = shift @apps;
	    }
	    elsif (s/.*\s$part1{$doing}(.*)$part2{$doing}.*/$1/) {
		$ver = $_;
		@tmp = split("$;", $found_versions{$doing});
		push(@tmp, $ver);
		$found_versions{$doing} = join("$;", @tmp);
	    }
	}
    }
    print LOG "Session to $site ended.\n";
    close LOG;
    unlink($logfile);
    return %found_versions;
}


sub source_url # ($app, $version, [$master])
#
# find an URL for the source to APP version VERSION on the store MASTER.
#
{
    local ($app, $ver, $mst) = @_;
    local ($src, $scheme, $user, $host, $port, $path);
    
    return undef unless ($app && $ver);
    unless (%url_scheme) {
	%url_scheme = ('http',1,
		       'ftp',1);
    }
    $src = &'getappinfo($app, 'source', $mst);
    $src =~ s/\$\{version\}/$ver/g;
    # see if the source field was an URL with a scheme we support
    if ($src =~ /^([a-z]+):/ && $url_scheme{$scheme = $1}) {
	$src =~ s/^[a-z]+://;
	if ($scheme eq 'http' || $scheme eq 'ftp') {
	    $host = $1 if $src =~ s=^/*([^:/]+)==;
	    $port = $1 if $src =~ s=^:(\d+)==;
	    $src = '/' if (!$src && $scheme eq 'http');
	    $path = $src;
	    return "$scheme://$host" . ($port ? ":$port" : '') . $path;
	}
    }
    undef;
}


sub add_files_for_checking # (@apps, $master)
#
# adds files to the spool assoc for all the applications in @apps on
# $master.
#
# the returned value is somewhat special, it is an array containing
# the names of applications that have source fields that are either
# not using the FTP protocol or not containing "${version}".  these
# have not been added to the spool assoc but should probably be
# downloaded and checked anyway.  it might be an application that
# simply does not have a version number in the file it is distributed
# in, but rather in some source file.  a mechanism will be added to
# let you customize how to figure out version numbers in such cases.
#
{
    local (@apps, $master) = @_;
    local ($scheme, $host, $path, $wildfile, $src, @always);
    
    foreach $app (@apps) {
	next unless $app;
	&'output('Trace', "checking $app\n", $$);
	$src = &'getappinfo($app, 'source', $opt_m);
	&'output('Debug', "Source field says \"$src\"\n", $$);
	##
	## try to recognize the source field.
	## recognizes a HTTP URL:
	if ($src =~ m=^(http)://=) {
	    $scheme = $1;
	}
	##
	## recognizes ftp://host/path and host:/path:
	elsif ($src =~ m=^ftp://([^/]+)(/.*)= || $src =~ m=^([^:]+):(.*)=) {
	    ($host, $path) = ($1, $2);
	    $scheme = 'ftp';
	}
	##
	## if the source field is empty, don't complain.
	elsif ($src =~ /^\s*$/) {
	    next;
	}
	if (!$scheme) {
	    &'output('Warning',
		     "Could not grok the source field of $app: \"$src\"\n",
		     $$);
	    next;
	}
	##
	## source that is available only with some other scheme than
	## FTP or source fields without wildcards get stuffed into an
	## array that is returned so they can always be recommended
	## for downloading/checking (phew :-).
	if ($scheme ne 'ftp' || $path !~ /\${version}/) {
	    push(@always, $app);
	    next;
	}
	##
	## turn "${anything}" into "*"
	$wildfile = $path;
	$wildfile =~ s/\$\{.+\}/\*/g ;
	&add_file_for_checking($app, $host, $wildfile);
    }
    @always;
}


sub handle_one_site # ($site, [$fork])
{
    local ($site, $fork) = @_;
    local ($pid, %found, @vers);
    
    if (!$fork || !($pid = fork)) {
	if ($fork) {
	    $SIG{'ALRM'} = 'exit';
	    alarm 10 * 60;
	    undef %sites_by_pids;
	    undef %apps_by_sites;
	    undef @sitequeue;
	    
	}
        &'output('Trace', "processing $site.\n", $$);
        %found = &check_site($site);
	while (($app, $versions) = each %found) {
	    @foundvers = split("$;", $versions);
	    @foundvers = sort { &'cmp_version($a, $b) } @foundvers;
	    &'output('Debug', "  $app: found @foundvers\n", $$);
	    @oldvers = split(/\s+/, &'getappinfo($app, 'versions', $opt_m));
	    @oldvers = sort { &'cmp_version($a, $b) } @oldvers;
	    
	    &'output('Debug', "  $app: had @oldvers\n", $$);
	    $rec = 0;
	    if (@oldvers) {
		foreach $found (@foundvers) {
		    foreach $old (@oldvers) {
			##
			## newer becomes: 0 -- the versions are equal,
			##                1 -- $found is newer than $old
			##               -1 -- $found is older than $old
			$newer = &'cmp_version($old, $found);
			##
			## if the found version is newer...
			if ($newer == 1) {
			    &'output('Dump',
				     "   $found is newer than $old\n", $$);
			    ##
			    ## ...if the user wants all versions, stack
			    ## them up for now.
			    if ($opt_v) {
				push(@vers, $found);
			    }
			    ##
			    ## ... if the user wants only the newest version,
			    ## compare the found version to our remembered
			    ## version.
			    else {
				## if the found one is newer, remember it.
				if (&'cmp_version($rec, $found) == 1) {
				    $rec = $found;
				}
			    }
			}
			elsif ($newer == -1) {
			    &'output('Dump',
				     "   $found is older than $old\n", $$);
			}
			else {
			    &'output('Dump', "   $found == $old\n", $$);
			}
			last if ($newer != 1);
		    }
		}
	    }
	    else {
		## if we had no versions already, pick the latest version
		## found (or all of them if -v is specified).
		if ($opt_v) {
		    @vers = @foundvers;
		}
		else {
		    $rec = $foundvers[$[];
		}
	    }
	    ##
	    ## if we've found a newer version, recommend it now.
	    if (@vers) {
		&recommend($app, 0, @vers);
	    }
	    elsif ($rec) {
		&'output('Debug',
			 " The last version of $app seems to be \"$rec\".\n",
			 $$);
		&recommend($app, $rec);
	    }
	    else {
		&'output('Debug',
			 " No newer versions of $app found.\n",
			 $$);
	    }
	}
	## we are the child process.
	return 0 if $fork;
	## we are not forking at all.
	return undef;
    }
    else {
        &'output('Debug', "forked off process $pid to handle $site,\n", $$);
        $site_procs{$pid} = $site;
        $parent = 1;
	return $pid;
    }
}
