# This is a -*- perl -*- package.
#
package usage;

$blocksize = 1024;  # Should really be dependent upon file system...

sub newspec
{
    local ( $spec ) = @_;

    $spec2speci{$spec}= $freespeci;
    $speci2spec[$freespeci]=$spec;
    $speci2files[$freespeci]=0;
    $speci2links[$freespeci]=0;
    $speci2size[$freespeci]=0;
    $freespeci++;
}

sub procapp
{
    local ( $sto, $app ) = @_;
    local ( 
	   $ver,
	   %files, 
	   %sumver,
	   $sumver,
	   $size,
	   %inomap,
	   %linkmap,
	   @usage,
	   %spec2speci,
	   @speci2spec,
	   @speci2files,
	   @speci2links,
	   @speci2size,
	   $freespeci,
	   $speci,
	   $spec,
	   $nspeci,
	   $nspec,
	   $ospeci,
	   $ospec
	   );
	    

    &'readregist($sto, $app );
    &'multi_version_parse() ;
    for $ver ( sort keys %'multisupport ) {

        # The files should be indexed both by inode/device, and by link name.

	undef %files;		# Array passed to summarize.
	undef %inomap;		# Map from device/inode to spec index.
	undef %linkmap;		# Map from link name to spec index.
	undef %spec2speci;	# Map from spec to spec index
	undef @speci2spec;	# Map from spec index to spec
	undef @speci2files;	# Map from spec index to file count.
	undef @speci2links;	# Map from spec index to file count.
	undef @speci2size;	# Map from spec index to file size.

	$freespeci=0;		# First free spec index.

	&'summarize( $sto, $app,  $ver, 0, *files);
	while ( ( $name, $attr ) = each %files ) {
	    ( $clen, $ctime, $cmode, $clink, $cdev, $cino ) = split(' ',$attr) ; 
	    if ( $name =~ m~^([^\@]*)\@([^\@/:,]*)$~ ) {
		$spec="arch:$2,domain:alldomains";
		$sf = $1 ;
	    } elsif ( $name =~ m~\@~ ) {
		# Note: genarchs'getarch need $sto/$app/$ver 
		( $sf, $d, $a, $l, $dc ) = &genarchs'getarch($name);
		if ( $l ne '' && $l =~ m=^(.*,|)linktree:([^,/]*)(,.*|)$= ) {
		    $spec="arch:$a,domain:$d,linktree:$2";
		} else {
		    $spec="arch:$a,domain:$d";
		}
	    }
	    else {
		$spec='arch:allarchs,domain:alldomains';
                $sf = $name ; 
	    }

            # Handling by inode/device 
	    &newspec($spec) unless ( defined $spec2speci{$spec} );
	    $speci=$spec2speci{$spec};
	    if ( defined $inomap{$cdev,$cino} ) {
		# Hard link.
		$ospeci = $inomap{$cdev,$cino};
		next if ( $ospeci == $speci );
		undef $ospec;
		for $ospec ( split('/', $speci2spec[$ospeci] ) ) {
                    last if $ospec eq $spec;
		}
		next if $ospec eq $spec;
		$nspec = join('/', 
			      sort(split('/',$speci2spec[$ospeci]), $spec));
		&newspec($nspec) unless ( defined $spec2speci{$nspec} );
		$nspeci = $spec2speci{$nspec};
		# Update number of files/ sizes.
		$speci2files[$ospeci]--;
		$speci2files[$nspeci]++;
		$size = int( ( $clen + ( $blocksize - 1 ) ) / $blocksize) ;
		$speci2size[$ospeci] -= $size ;
		$speci2size[$nspeci] += $size;
		$inomap{$cdev,$cino} = $nspeci; 
	    } else {
		# No hard link detected yet. 
		$inomap{$cdev,$cino} = $speci ; 
		$size = int( ( $clen + ( $blocksize - 1 ) ) / $blocksize) ;
		$speci2size[$speci]  += $size;
		$speci2files[$speci]++;
	    }

            # Handling by link name 

	    if ( defined $linkmap{$sf} ) {
		# Multiple occurences of same link name 
		$ospeci = $linkmap{$sf};
		next if ( $ospeci == $speci ); # Same spec. Only one used.
		undef $ospec;
		for $ospec ( split('/', $speci2spec[$ospeci] ) ) {
                    last if $ospec eq $spec;
		}
		next if $ospec eq $spec; # Same spec already present.
		$nspec = join('/', 
			      sort(split('/',$speci2spec[$ospeci]), $spec));
		&newspec($nspec) unless ( defined $spec2speci{$nspec} );
		$nspeci = $spec2speci{$nspec};
		# Update number of files/ sizes.
 		$speci2links[$ospeci]--;
		$speci2links[$nspeci]++;
		$linkmap{$sf} = $nspeci; 
	    } else {
		# First occurence of this link.
		$linkmap{$sf} = $speci; 
		$speci2links[$speci]++;
	    }
	}
	for ( $speci=0;$speci<$freespeci;$speci++) {
	    $size = int( ($speci2size[$speci] * $blocksize + 1023 ) / 1024 );
	    $files = $speci2files[$speci];
	    $links = $speci2links[$speci];
	    $spec = $speci2spec[$speci];
	    next unless ( $files || $links || $size ); 
	    push(@usage, "$app\@$sto $ver $spec $links $files $size");
	}
    }
    return @usage;
}


sub process
{
    local ( $sto, $app ) = @_;
    local ( $sapp );
    local ( @usage );
    local ( @applist );

    if ( $app ) {
	@applist = ( $app );
    } else {
	@applist = &'getapplist($sto);
    }
    
    for $app ( @applist ) {
	if ( -d "$'storedir{$sto}/$app" ) {
	    $sapp = $app ;
	} elsif ( -d "$'storedir{$sto}/.$app" ) {
	    $sapp = ".$app" ;
	} else {
	    &'output('Trace','Application not found','Usage',$sto,$app);
	    next;
	}
	&'output('Trace',"Processing\n",'Usage',$sto,$sapp);
	push(@usage, &procapp($sto,$sapp));
    }
    return @usage;
}

sub getinfo {
    local( $store, $app , $force) = @_;
    local ( @usage );
    local ( $cachefile, $ncachefile );

    $cachefile = "$'storedir{$store}/store/usage";
    $ncachefile = "$'storedir{$store}/store/usage.new";
    # 
    # Read information from a cached file.
    # If $force is set, or the cache file
    # doesn't exist, it is created.
    #
    if ( $force || ! -r $cachefile ) {
	&'output('Trace',"Store not available\n",'Usage',$store),return
	    unless ( -d $'storedir{$store} );
	@usage = &process($store, $app );
	# Write cache information 
	if ( ! $app &&  $'master{$store} eq $'myself ) {
          &'createdir($ncachefile );
          open(CACHEFILE,">$ncachefile");
          for $one ( @usage ) {
	      print CACHEFILE $one,"\n";
	  }
	  close(CACHEFILE);
	  rename($ncachefile,$cachefile);
      }
    } else {
	open(CACHEFILE,$cachefile);
	if ( $app ne '') {
	    $sapp = ".$app" ;
	    while ( <CACHEFILE> ) {
		chop;
		push(@usage,$_)
		    if ( m/^([^\@]*)\@/ && ( $1 eq $app || $1 eq $sapp ) );
	    }
	} else {
	    while ( <CACHEFILE> ) {
		chop;
		push(@usage,$_);
	    }
	}
	close(CACHEFILE);
    }
    return @usage;
}

1;
